/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2002-2003 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

static const char* const MRtpSession_cxx_Version =
    "$Id: MRtpSession.cxx,v 1.4.2.3 2003/03/13 17:57:35 bko Exp $";

#include "global.h"
#include <cassert>
#include "RtpPacket.hxx"

#include "MediaDevice.hxx"
#include "MediaSession.hxx"
#include "MRtpSession.hxx"
#include "MediaController.hxx"

using namespace Vocal::UA;
using namespace Vocal::SDP;


MRtpSession::MRtpSession(int sessionId, NetworkRes& local, NetworkRes& remote ,
                         Sptr<CodecAdaptor> cAdp, u_int32_t ssrc)
            : Adaptor(RTP, NONE), mySessionId(sessionId), myShutdown(false)
{
    myCodec = cAdp;
    myLocalAddress = new NetworkRes(local);
    myRemoteAddress = new NetworkRes(remote);
    int remotePort = myRemoteAddress->getPort();
    int localPort  = myLocalAddress->getPort();
    int remoteRtcpPort = ( remotePort > 0) ? remotePort + 1 : 0;
    int localRtcpPort = (localPort > 0) ? localPort + 1 : 0;

    RtpPayloadType pType;
    switch(cAdp->getType())
    {
        case G711U:
            pType = rtpPayloadPCMU;
        break;
        case G711A:
            pType = rtpPayloadPCMA;
        break;
        case G729:
            pType = rtpPayloadG729;
        break;
        default:
            pType = rtpPayloadPCMU;
        break;  
    }

    rtpStack =  new RtpSession( myRemoteAddress->getIpName().c_str(), 
                                remotePort,
                                localPort,
                                remoteRtcpPort,
                                localRtcpPort, pType,
                                pType, 0 );
    myDTMFInterface = new MDTMFInterface(this);
    rtpStack->setDTMFInterface(myDTMFInterface);
    rtpStack->setMarkerOnce();
    threadStarted = false;
}

void
MRtpSession::start()
{
    if(!threadStarted)
    {
        if((rtpStack->getSessionState() == rtp_session_recvonly) ||
           (rtpStack->getSessionState() == rtp_session_sendrecv))
        {
            myThread.spawn(processThreadWrapper, this); 
            threadStarted = true;
        }
    }
}

void
MRtpSession::processIncomingRTP()
{
    int incomingFd = -1;

    myMutex.lock();
    if ( myShutdown == true)
    {
         myMutex.unlock();
         return;
    }
    myMutex.unlock();

    RtpReceiver* recv = rtpStack->getRtpRecv();
    if(recv)
    {
        incomingFd = recv->getSocketFD();
    }

    while(incomingFd > 0)
    {

        fd_set netFD;
        FD_ZERO (&netFD);
        FD_SET(incomingFd, &netFD);

        struct timeval timeout;
        timeout.tv_sec = 0;
        timeout.tv_usec = 0;

        int selret = select(incomingFd + 1, &netFD, 0, 0, &timeout);

        if(selret > 0)
        {
            RtpSessionState sessionState = rtpStack->getSessionState();
            
            if ( sessionState == rtp_session_undefined ) 
            {
                cpLog(LOG_ERR, "Session state undefined, returning");
                return;
            }
            
            // receive audio
            if( sessionState == rtp_session_recvonly ||
                sessionState == rtp_session_sendrecv )
            {
                RtpPacket* packet = rtpStack->receive();
                int nRtpRate = myDataRate * (myCodec->getClockRate()/ 1000);
                if( packet )
                {
                    if( packet->getPayloadType() == (RtpPayloadType)myCodec->getType() &&
                        packet->getPayloadUsage() == nRtpRate )
                    {
                        processRecv(packet, *myRemoteAddress);
                    }
                    else
                    {
                        cpLog(LOG_ERR,
                              "Received from RTP stack,incorrect payload type");
                    }
                    delete packet; 
                }
            }
            else
            {
                return;
            }
        }
        else
        {
            return;
        }
    }
}

void* 
MRtpSession::processThreadWrapper(void *p)
{
    MRtpSession* self = static_cast<MRtpSession*>(p);
    self->thread();
    return 0;
}


void
MRtpSession::thread()
{
    threadStarted = true;
    while ( true )
    {
        myMutex.lock();
        if ( myShutdown == true)
        {
             myMutex.unlock();
             mySession = 0;
             break;
        }
        myMutex.unlock();

        int incomingFd=-1;
        RtpReceiver* recv = rtpStack->getRtpRecv();
        if(recv)
        {
            incomingFd = recv->getSocketFD();
        }

        fd_set netFD;
        FD_ZERO (&netFD);
        if(incomingFd > 0)
        {
            FD_SET(incomingFd, &netFD);
        }

        int maxfd = incomingFd;

        struct timeval timeout;
        timeout.tv_sec = 1;
        timeout.tv_usec = 0;

        int selret = select(maxfd + 1, &netFD, 0, 0, &timeout);

        if(selret > 0)
        {
            processIncomingRTP();
        }
    }
    cpLog(LOG_DEBUG, "MRtpSession::thread-exiting...");
    return;
}

void 
MRtpSession::processRecv(RtpPacket* packet, const NetworkRes& sentBy)
{
    if(myShutdown == true)
    {
        mySession = 0;
        return;
    }

    //Cache the session obj
    if(mySession == 0)
    {
        mySession = MediaController::instance().getSession(mySessionId);
    }
    mySession->processRaw((char*) packet->getPayloadLoc(), packet->getPayloadUsage(), myCodec->getType(), this);
}

///Consume the raw data
void 
MRtpSession::sinkData(char* data, int length, VCodecType type)
{
    if(myShutdown == true)
    {
        mySession = 0;
        return;
    }

    //if the codec type is not same as the network type of the session
    //convert back, say for example if data is PCMU and sent over session codec
    //is G729, convert data first to G729. If G729 data is coming as input, no converssion
    //is needed and send it as is
    //Send RTP packet to the destination
    cpLog(LOG_DEBUG_STACK, "%s sinking data", description().c_str());
    int retVal = 0;
    if(type == DTMF_TONE)
    {
        //First 8 bits will represent the event
        char event;
        memcpy(&event, data, 1);
        cerr << "Sending digit:" << (int)event << endl;
        retVal = rtpStack->transmitEvent(event);
    }
    else 
    {
        if(type != myCodec->getType())
        {
           //DO the necessary conversion here
        }
        cpLog(LOG_DEBUG_STACK, "Transmit raw:%d", length);
        retVal = rtpStack->transmitRaw(data, length);
    }

    if(retVal < 0)
    {
        cpLog(LOG_ERR, "Failed to transmit RTP packets to %s:%d", 
                       myRemoteAddress->getIpName().c_str(),
                       myRemoteAddress->getPort());
    }
}

string
MRtpSession::description()
{
    char buf[256];
    sprintf(buf, "%s:%d->%s:%d", myLocalAddress->getIpName().c_str(),
                                 myLocalAddress->getPort(),
                                 myRemoteAddress->getIpName().c_str(),
                                 myRemoteAddress->getPort());
    return buf;
}

void 
MRtpSession::recvDTMF(int event)
{
    cerr << "Received:" << event << endl;
    char eventData = event; 
    MediaController::instance().getSession(mySessionId)->processRaw(&eventData, 1, DTMF_TONE, this);
}

void 
MRtpSession::adopt(SdpSession& remoteSdp)
{
    Data rAddress;
    int rPort;
    //Check for address in Media part if found find else use the
    //Global address
    list<SdpMedia*> mediaList = remoteSdp.getMediaList();
    for(list<SdpMedia*>::iterator itr = mediaList.begin(); itr != mediaList.end(); itr++)
    {
        if((*itr)->getMediaType() == Vocal::SDP::MediaTypeAudio)
        {
            if((*itr)->getConnection())
            {
                cpLog(LOG_DEBUG, "Connection address found in media attribute, using it");
                rAddress = (*itr)->getConnection()->getUnicast();
                break;
            }
            rPort = (*itr)->getPort();
        }
    }

    if(rAddress.length() == 0)
    {
        //Check to see if Session-level connection address is present
        SdpConnection* conn = remoteSdp.getConnection();
        if(conn) rAddress = conn->getUnicast();
        if(rAddress.length())
        {
            cpLog(LOG_DEBUG, "Using Top-level connection address");
        }
        else
        {
            cpLog(LOG_ERR, "Bad SDP");
        }
    }
    myRemoteAddress->setHostName(rAddress);
    myRemoteAddress->setPort(rPort);
}

void
MRtpSession::setMode(VSdpMode mode)
{
    switch(mode)
    {
        case VSDP_SND:
            rtpStack->setSessionState(rtp_session_sendonly);
        break;
        case VSDP_RECV:
            rtpStack->setSessionState(rtp_session_recvonly);
        break;
        case VSDP_SND_RECV:
            rtpStack->setSessionState(rtp_session_sendrecv);
        break;
    }
}


/* Local Variables: */
/* c-file-style: "stroustrup" */
/* indent-tabs-mode: nil */
/* c-file-offsets: ((access-label . -) (inclass . ++)) */
/* c-basic-offset: 4 */
/* End: */
