/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */


static const char* const MediaController_cxx_Version = 
    "$Id: MediaController.cxx,v 1.3 2002/12/20 23:53:37 sprajpat Exp $";


#include "MediaController.hxx"
#include "SdpHandler.hxx"
#include "Lock.hxx"
#include "CodecG711U.hxx"
#include "CodecTelEvent.hxx"

using namespace Vocal;
using namespace Vocal::UA;

MediaController* MediaController::myInstance = 0;

MediaController&
MediaController::instance()
{
    assert(myInstance != 0);
    return *myInstance;
}

MediaController&
MediaController::instance(int minRtpPort, int maxRtpPort)
{
    assert(myInstance == 0);
    if(myInstance == 0)
    {
        myInstance = new MediaController(minRtpPort, maxRtpPort);
    }
    return *myInstance;
}

MediaController::MediaController(int minRtpPort, int maxRtpPort)
{
    cpLog(LOG_DEBUG, "MediaController::MediaController");
    myRollingSessionId = 1;
    //get the list of reserved ports
    for(int i = minRtpPort; i < maxRtpPort; i = i+2)
    {
        //Check to see if port is free
        try
        {
            UdpStack uStack(0, i , i );
            Sptr<NetworkRes> res = new NetworkRes(i);
            myNetworkResList.push_back(res);
        }
        catch(...)
        {
             cpLog(LOG_ERR, "Port %d is busy, try next", i);
        }
    } 

    ///Create the list of supported codec, default list
    Sptr<CodecAdaptor> cAdp = new CodecG711U();
    registerCodec(cAdp);

    Sptr<CodecAdaptor> cAdp2 = new CodecTelEvent();
    registerCodec(cAdp2);

    cpLog(LOG_DEBUG, "Initialized MediaController");
}

MediaController::~MediaController()
{
}

void
MediaController::shutdown()
{
}


SdpSession 
MediaController::createSession(const SdpSession& remoteSdp) 
    throw (MediaException&) 
{
    //Based on the remote SDP and capability, create a
    //compatible Local SDP and reserve associated Media
    //Resources 
    //reserve a local audio device
    string localAddr;
    int    localPort;
    
    int sId = createSessionImpl( localAddr, localPort);
    SdpSession sdp;
    negotiateSdp(sdp, localAddr, localPort, remoteSdp);
    sdp.setSessionId(sId);

    return sdp;
}

SdpSession 
MediaController::createSession() throw (MediaException&)
{
    string localAddr;
    int    localPort;
    
    int sId = createSessionImpl( localAddr, localPort);
    SdpSession sdp;
    setStandardSdp(sdp, localAddr, localPort);
    sdp.setSessionId(sId);
    return sdp;
}

int
MediaController::createSessionImpl(string& localAddr, int& port)
{
    //Get a NetworkRes from the list of resources and devices
    //and create a session Id to assign
    Lock lock(myMutex);
    int sId = myRollingSessionId++;
    //Get a free NetworkRes and a device
    Sptr<NetworkRes> localRes; 
    for(NetworkResList::iterator itr = myNetworkResList.begin();
            itr != myNetworkResList.end(); itr++)
    {
        if(!(*itr)->isBusy())
        {
            localRes = *itr;
            localRes->setBusy(true);
            localAddr = localRes->getIpName().c_str();
            port = localRes->getPort();
            break;
        }
    }
    if(localRes == 0)
    {
        cpLog(LOG_ERR, "No Network resource is free");
        throw MediaException("No Network resource is free",
                             __FILE__, __LINE__);
    }

    Sptr<MediaSession> mSession = new MediaSession(sId, localRes);

    assert(myMediaSessionMap.count(sId) == 0);
    myMediaSessionMap[sId] = mSession;
    return sId;
}

void 
MediaController::freeSession(int sId)
{
    //Free the resources associated with the session Id
    Lock lock(myMutex);
    MediaSessionMap::iterator itr = myMediaSessionMap.find(sId);
    if(itr == myMediaSessionMap.end())
    {
        cpLog(LOG_DEBUG, "Session %d does not exists", sId);
        return;
    }
    if((itr->second)->tearDown())
    {
         cpLog(LOG_DEBUG, "Freeing session %d", sId);
         myMediaSessionMap.erase(itr);
    }
}

void
MediaController::addDeviceToSession(unsigned int sessionId, Sptr<MediaDevice> mDevice)
{
    Lock lock(myMutex);
    if(myMediaSessionMap.count(sessionId) == 0)
    {
        cpLog(LOG_ERR, "No session exists for if (%d), ignoring the add device request",
                        sessionId);
        return;
    }
    Sptr<MediaSession> mSession = myMediaSessionMap[sessionId];
    mSession->addToSession(mDevice);
}

void
MediaController::addToSession(SdpSession& localSdp, SdpSession& remoteSdp)
{
    unsigned int sId = localSdp.getSessionId();
    cpLog(LOG_DEBUG, "Adding to session (%d)", sId);
    Lock lock(myMutex);
    assert(myMediaSessionMap.count(sId));
    Sptr<MediaSession> mSession = myMediaSessionMap[sId];
    mSession->addToSession(localSdp, remoteSdp);
}

Sptr<MediaSession>
MediaController::getSession(unsigned int sessionId)
{
    Lock lock(myMutex);
    MediaSessionMap::iterator itr = myMediaSessionMap.find(sessionId);
    assert(itr != myMediaSessionMap.end());
    return((*itr).second);
}

void 
MediaController::registerDevice(Sptr<MediaDevice> mDevice)
{
   myMediaDeviceList.push_back(mDevice);
}

list<Sptr<MediaDevice> >
MediaController::getListOfMediaDevices()
{
    //get the list fo devices audio/video which are currently
    //available
    list<Sptr<MediaDevice> > retList;
    //Return list of devices which either support audio or video
    for(MediaDeviceList::iterator itr = myMediaDeviceList.begin();
         itr != myMediaDeviceList.end(); itr++)
    {
        if((*itr)->getMediaType() != NONE ) 
        {
            Sptr<MediaDevice> mDevice;
            mDevice.dynamicCast((*itr));
            if(!mDevice->isBusy())
            {
                retList.push_back(mDevice);
            }
        }
    }
    return retList;
}


void 
MediaController::registerCodec(Sptr<CodecAdaptor> cAdp)
{
    myMediaCapability.addCodec(cAdp);
}

void
MediaController::startSession(unsigned int sId, VSdpMode mode)
{
    getSession(sId)->startSession(mode);
}

void
MediaController::suspendSession(unsigned int sId)
{
    getSession(sId)->suspend();
}
void
MediaController::resumeSession(unsigned int sId, SdpSession& remoteSdp)
{
    getSession(sId)->resume(remoteSdp);
}

SdpSession
MediaController::getSdp(unsigned int sId, VSdpMode mode)
{
    return (getSession(sId)->getSdp(mode));
}

SdpSession
MediaController::getSdp(unsigned int sId, SdpSession& remoteSdp)
{
    return (getSession(sId)->getSdp(remoteSdp));
}


