#ifndef MediaController_hxx
#define MediaController_hxx

/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */


static const char* const MediaController_hxx_Version = 
    "$Id: MediaController.hxx,v 1.1.2.1 2003/01/28 02:16:56 sprajpat Exp $";

#include "global.h"
#include <list>
#include <map>
#include "Sptr.hxx"
#include "Rtp.hxx"
#include "VMutex.h"
#include "VException.hxx"
#include "MediaDevice.hxx"
#include "NetworkRes.hxx"
#include "MediaSession.hxx"
#include "Sdp2Session.hxx"
#include "MediaCapability.hxx"
#include "MediaException.hxx"

using Vocal::SDP::SdpSession;

namespace Vocal
{

namespace UA
{


/** MediaController is a singleton object to manage Media session between two or more
    parties. The Controller provide interfaces for managing various
    media devices and allocate network resources for a given session. 

<pre>
    app_main()
    {
        int myMinRtpPort = 10000, maxRtpPort = 12000;
        MediaController::instance(10000, 12000);
   
        //offering a SDP (UAC)
        SdpSession localSdp = MediaController::instance().createSession(); 
        unsigned sessionId = localSdp.getSessionId();
        //save the sessionId for later use
        //send the request with localSdp 
        //wait for response  
        ...
        ...
        //When the answer comes with remoteSdp
        MediaController::instance().addToSession(localSdp, remoteSdp);

        Sptr<MediaDevice> mDevice = //Create a media device
        MediaController::instance().addDeviceToSession(sessionId, mDevice);
        MediaController::instance().startSession(sessionId); 
        //Session started
        ...
        ...

        //At the end of the session
        MediaController::instance().freeSession(sessionId);
    }
</pre>
 */
class MediaController
{
    public:
      ///Interface to access the Singelton object.
      static MediaController& instance();

      /**Interface to initialize the Controller. The interface should be called
        *by the application only once to create and intialize the Controller object.
        */
      static MediaController& instance(int minRtpPort, int maxRtpPort);

      ///
      string className() { return "MediaController"; }

      /** Virtual destructor
       */
      virtual ~MediaController();

      ///
      typedef list<Sptr<Adaptor> > MediaDeviceList;

      ///
      typedef list<Sptr<NetworkRes> > NetworkResList;

      ///
      typedef map<int , Sptr<MediaSession> > MediaSessionMap;


      ///Create a compatibl;e session after negotiating with remote SDP
      SdpSession createSession(const SdpSession& remoteSdp)
             throw (MediaException&) ;

      /**Create a brand new SDP session
        * using the RTP port range and MediaCapabilities registered.
        */
      SdpSession createSession()
             throw (MediaException&);

      ///Get SDP for an existing session identified by sId.
      SdpSession getSdp(unsigned int sId, VSdpMode=VSDP_SND_RECV);

      /**Get SDP for an existing session identified by sId, after negotiatng
        * with given remoteSdp.
        */
      SdpSession getSdp(unsigned int sId, SdpSession& resmoteSdp);

      ///Free a media session and all its resources identified by sId.
      void freeSession(int sessionId);

      ///Join a remote site to a session identified by localSdp 
      void addToSession(SdpSession& localSdp, SdpSession& remoteSdp);

      ///Add a device to session identified by the sessionId
      void addDeviceToSession(unsigned int sessionId, Sptr<MediaDevice> mDevice);

      ///
      void shutdown();

      ///Register a media device
      void registerDevice(Sptr<MediaDevice> mDevice);

      ///Register a Codec 
      void registerCodec(Sptr<CodecAdaptor> cAdp);

      ///Get the list of media devices registered with the controller
      list<Sptr<MediaDevice> > getListOfMediaDevices();

      ///Get the MediaCapability object that holds the list of codecs supported
      MediaCapability& getMediaCapability() { return myMediaCapability; };

      ///Get a session associated with a sessionId.
      Sptr<MediaSession> getSession(unsigned int sessionId);

      ///Starts a session identified by sId and given mode.
      void startSession(unsigned int sId, VSdpMode mode=VSDP_SND_RECV);

      ///
      void suspendSession(unsigned int sId);

      ///
      void resumeSession(unsigned int sId, SdpSession& remoteSdp);

    private:
        ///
        int createSessionImpl(string& localAddr, int& port);

        /** Create the Instance of MediaController
         */
        MediaController(int minRtpPort, int maxRtpPort);

        ///
        static MediaController* myInstance;

        ///List of media devices managed by the Media Controller
        MediaDeviceList  myMediaDeviceList;
        /**List of NetworkResource objects ( encapsulates IP,Port 
         * pair of listening
         * sending media, and also the availability information)
         */
        NetworkResList myNetworkResList;

        ///
        MediaSessionMap myMediaSessionMap;

        /** Suppress copying
         */
        MediaController(const MediaController &);
        
        /** Suppress copying
         */
        const MediaController & operator=(const MediaController &);
        ///
        int   myRollingSessionId;
        ///
        VMutex   myMutex;

        ///
        MediaCapability myMediaCapability;
};

}
}

#endif
