
/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

static const char* const MediaSession_cxx_Version =
    "$Id: MediaSession.cxx,v 1.4 2002/12/20 23:53:37 sprajpat Exp $";

#include "global.h"
#include <cassert>

#include "MediaDevice.hxx"
#include "MediaSession.hxx"
#include "Sdp2Session.hxx"
#include "MediaController.hxx"
#include "SdpHandler.hxx"

#include "MRtpSession.hxx"

using namespace Vocal::UA;
using namespace Vocal::SDP;


MediaSession::MediaSession(int sessionId, 
                           Sptr<NetworkRes> localRes) 
    : mySessionId(sessionId)
{
    mySSRC = 0;
    while(mySSRC == 0) mySSRC =  rand();
    localRes->setBusy(true);
    myLocalRes = localRes;
}


MediaSession::~MediaSession()
{
    cpLog(LOG_DEBUG, "MediaSession::~MediaSession");
    myLocalRes->setBusy(false);
    if(myMediaDevice != 0)
    {
        myMediaDevice->setBusy(false);
        myMediaDevice->assignedTo(-1);
    }
}
 
void
MediaSession::addToSession( Sptr<MediaDevice> mDevice)
{
   if(mDevice != 0)
   {
       mDevice->setBusy(true);
       mDevice->assignedTo(mySessionId);
       myMediaDevice = mDevice;
   }
}


void
MediaSession::addToSession( SdpSession& localSdp, SdpSession& remoteSdp)
{
    if(myRtpSession != 0)
    {
        cpLog(LOG_ERR, "Already has a session, ignoring...");
        return;
    }
    // Set remote host and port
    //   For now only support unicast
    string remAddr;
    string localAddr;
    LocalScopeAllocator lo;
    if(localSdp.getConnection()) localAddr = localSdp.getConnection()->getUnicast().getData(lo);
    if(remoteSdp.getConnection()) remAddr = remoteSdp.getConnection()->getUnicast().getData(lo);

    //Get the Negotiated Codec for each media type
    //Get each media description from remoteSdp and map it to the
    //localSdp SDP and create a MRtpSession
    list < SdpMedia* > mList = remoteSdp.getMediaList();
    MediaController& mInstance = MediaController::instance();
    for(list < SdpMedia* >::iterator itr = mList.begin();
                            itr != mList.end(); itr++)
    {
        SdpMedia* rMedia = (*itr);
        string rAddr;
        string lAddr;
        int lPort;
        int rPort;
        Sptr<CodecAdaptor> cAdp;
        rPort = rMedia->getPort();
        list < SdpMedia* > lmList = localSdp.getMediaList(); 
        for(list < SdpMedia* >::iterator itr2 = lmList.begin();
                            itr2 != lmList.end(); itr2++)
        {
            SdpMedia* lMedia = (*itr2);
            if(lMedia->getMediaType() != rMedia->getMediaType())
            {
                continue;
            }
            //Get the very first codec of the media description
            vector <int>* fmList = lMedia->getFormatList();
            assert(fmList->size());
            int fmt = (*fmList)[0];
            cAdp = mInstance.getMediaCapability().getCodec(static_cast<VCodecType>(fmt)); 
            if(lMedia->getConnection())
            {
                LocalScopeAllocator lo;
                lAddr = lMedia->getConnection()->getUnicast().getData(lo);
            }
            else
            {
                lAddr = localAddr;
            }
            lPort = lMedia->getPort();
            break;
        }
        if(rMedia->getConnection())
        {
            LocalScopeAllocator lo;
            rAddr = rMedia->getConnection()->getUnicast().getData(lo);
        }
        else
        {
            rAddr = remAddr;
        }
        //Now if we have lAddr, rAddr, lport, rPort and codec adaptor
        if((lAddr.size() ==0) || (rAddr.size() == 0) || (cAdp == 0))
        {
           cpLog(LOG_ERR, "Media is not setup correctly");
           continue;
        }
        //Create a RTP session
        NetworkRes localRes(lAddr, lPort);
        NetworkRes remoteRes(rAddr, rPort);
        Sptr<MRtpSession> mSession = 
                                 new MRtpSession(mySessionId, localRes,
                                                 remoteRes,
                                                 cAdp, mySSRC);
        myRtpSession = mSession;
    }
}

int 
MediaSession::tearDown()
{
    cpLog(LOG_DEBUG, "Tearing down the session:%d" , mySessionId);
    if(myMediaDevice != 0)
    {
        myMediaDevice->stop();
        myMediaDevice->setBusy(false);
    }

    if(myRtpSession != 0) 
    {
        myRtpSession->shutdown();
        myRtpSession = 0;
    }

    myLocalRes->setBusy(false);
    return 1;
}

void 
MediaSession::processRaw(char *data, int len, VCodecType cType, Adaptor* adp)
{
    if(adp->getDeviceType() != RTP)
    {
        //Data from hardware, ship it out to the RTP session
        if(myRtpSession != 0) myRtpSession->sinkData(data, len, cType);
    }
    else
    {
         myMediaDevice->sinkData(data, len, cType);
    }
}

void
MediaSession::startSession(VSdpMode mode)
{
    myMediaDevice->start(); 
    Sptr<MRtpSession> mSession;
    mSession.dynamicCast(myRtpSession);
    if(mSession != 0)
    {
        mSession->setMode(mode); 
        mSession->start(); 
    }
}

void
MediaSession::suspend()
{
    myMediaDevice->suspend(); 
}
void
MediaSession::resume(SdpSession& remoteSdp)
{
    Sptr<MRtpSession> mSession;
    mSession.dynamicCast(myRtpSession);
    if(mSession != 0)
        mSession->adopt(remoteSdp);
    myMediaDevice->resume(); 
}

SdpSession 
MediaSession::getSdp(VSdpMode mode)
{
    string localAddr;
    int    localPort;
    localAddr = myLocalRes->getIpName().c_str();
    localPort = myLocalRes->getPort();
    SdpSession sdp;
    setStandardSdp(sdp, localAddr, localPort, mode);
    sdp.setSessionId(mySessionId);
    return sdp;
}
 

SdpSession 
MediaSession::getSdp(SdpSession& remoteSdp)
{
    string localAddr;
    int    localPort;
    localAddr = myLocalRes->getIpName().c_str();
    localPort = myLocalRes->getPort();
    SdpSession sdp;
    negotiateSdp(sdp, localAddr, localPort, remoteSdp);
    sdp.setSessionId(mySessionId);
    return sdp;
}
