/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

static const char* const OpStartRingbackTone_cxx_Version =
    "$Id: OpStartRingbackTone.cxx,v 1.37 2002/11/09 02:13:07 derekm Exp $";

#include "global.h"
#include <cassert>

#include "cpLog.h"
#include "SipEvent.hxx"
#include "StatusMsg.hxx"
#include "OpStartRingbackTone.hxx"
#include "UaHardwareEvent.hxx"
#include "UaDevice.hxx"
#include "UaCallInfo.hxx"
#include "UaConfiguration.hxx"
#include "SdpHandler.hxx"
#include "SipSdp.hxx"

#ifdef HAS_RSVPLIB
#include "SuaRsvp.hxx"
#endif


using namespace Vocal;


OpStartRingbackTone::OpStartRingbackTone()
{
}


OpStartRingbackTone::~OpStartRingbackTone()
{
}


const char* const
OpStartRingbackTone::name() const
{
    return "OpStartRingbackTone";
}


const Sptr < State >
OpStartRingbackTone::process( const Sptr < SipProxyEvent > event )
{
    cpLog( LOG_DEBUG, "OpStartRingbackTone operation" );

    Sptr < SipEvent > sipEvent;
    sipEvent.dynamicCast( event );
    if ( sipEvent == 0 )
    {
        return 0;
    }
    Sptr < SipMsg > sipMsg = sipEvent->getSipMsg();
    assert( sipMsg != 0 );

    Sptr < StatusMsg > msg;
    msg.dynamicCast( sipMsg );
    if ( msg == 0 )
    {
        return 0;
    }

    Sptr < UaCallInfo > call;
    call.dynamicCast( event->getCallInfo() );
    assert( call != 0 );

    if ( msg->getStatusLine().getStatusCode() != 180 &&
            msg->getStatusLine().getStatusCode() != 183 )
    {
        return 0;
    };

    if ( cancelTimer(event) )
    {
        cpLog( LOG_DEBUG, "Ua Timer cancelled" );
    }

    // This is a temporary fix to handle the case in which more than one
    // 180/183 messages are received before 200 arrives
    // TODO: check sdp in the 180/183.  If different from previous 180/183,
    // use the lastest 180/183 sdp
    Sptr < Contact > contact = call->findContact( *msg );
    assert( contact != 0 );

    int status = contact->getStatus();
    if ( status == 180 || status == 183 )
    {
        cpLog( LOG_ERR, "Have existing 180/183 message" );
        cpLog( LOG_ERR, "Igore the current 180/183 message" );
        return 0;
    }

    cpLog( LOG_DEBUG, "Start Ringback tone" );

    bool remoteRingback = true;

    //update contact
    contact->update( *msg );

    //figure out if this is the first call or the second call
    Sptr < SipSdp > localSdp;
    Sptr < InviteMsg > inviteMsg = call->getRingInvite();

    InviteMsg invMsg = call->getContact()->getInviteMsg();
    if ( *inviteMsg == invMsg )
    {
        localSdp = call->getLocalSdp();
    }
    else
    {
	cpLog(LOG_DEBUG, "INVITE did not match, getting Local2 SDP");
	cpLog(LOG_DEBUG_STACK, 
	      "Original: %s", 
	      (inviteMsg)->encode().logData());
	cpLog(LOG_DEBUG_STACK, "New: %s", invMsg.encode().logData());
        localSdp = call->getLocal2Sdp();
    }

    /// should this really be an assert?  this seems dangerous
    assert( localSdp != 0 );
    
    // Set rtpPacketSize to default in ua.cfg file
    int rtpPacketSize = UaConfiguration::instance()->getNetworkRtpRate();

    Sptr<SipSdp> remoteSdp;
    remoteSdp.dynamicCast( sipMsg->getContentData(0) );
    
    if( remoteSdp == 0 )
    {
        cpLog( LOG_DEBUG, "No remote SDP in 180/183" );
        remoteRingback = false;
    }
    else
    {
	// get the remote SDP's rtp rate
	int rtpPacketSize = getRtpPacketSize(*remoteSdp);

	if(rtpPacketSize > 0)
	{
	    setRtpPacketSize(*localSdp, rtpPacketSize);
	    call->setRemoteSdp( new SipSdp( *remoteSdp ) );
	}
	else
	{
	    remoteRingback = false;
	}
    }

    Sptr < UaHardwareEvent > signal =
	new UaHardwareEvent( UaDevice::getDeviceQueue() );
    
    if ( remoteRingback )
    {
        cpLog( LOG_DEBUG, "Early RTP media" );

        call->getContact()->setRemoteRingback( true );

        signal->type = HardwareAudioType;

        struct HardwareAudioRequest* request 
	    = &(signal->signalOrRequest.request);

        request->type = AudioStart;

        strcpy( request->remoteHost, "\0" );
        request->remotePort = 0;
        LocalScopeAllocator lo;
        strcpy( request->localHost, localSdp->getConnAddress().getData(lo) );
        request->localPort = localSdp->getRtpPort();

//        request->echoCancellation = true;

        request->rtpPacketSize = rtpPacketSize;
    }
    else
    {
        call->getContact()->setRemoteRingback( false );
        signal->type = HardwareSignalType;
        signal->signalOrRequest.signal = DeviceSignalLocalRingbackStart;
    }

    UaDevice::getDeviceQueue()->add( signal );

    return 0;
}


/* Local Variables: */
/* c-file-style: "stroustrup" */
/* indent-tabs-mode: nil */
/* c-file-offsets: ((access-label . -) (inclass . ++)) */
/* c-basic-offset: 4 */
/* End: */
